<?php

namespace HS\IncomingMail\Jobs;

use HS\Mailbox;
use HS\Cloud\IsHosted;
use HS\Jobs\AdministratesJobs;
use HS\IncomingMail\Mailboxes\Imap;

use Illuminate\Support\Str;
use Illuminate\Bus\Queueable;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Cache;
use Illuminate\Queue\SerializesModels;
use Illuminate\Support\Facades\Storage;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;

class FetchMail extends MailJobBase implements ShouldQueue
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels, AdministratesJobs, IsHosted;

    protected $jobName = 'Fetching New Emails';
    protected $jobCategory = 'mail.incoming';

    /**
     * @var Mailbox
     */
    private $mailbox;

    /**
     * @var string
     */
    private $transaction;

    /**
     * The maximum number of exceptions to allow before failing.
     *
     * @var int
     */
    public $maxExceptions = 1;

    /**
     * Create a new job instance.
     *
     * @param Mailbox $mailbox
     * @param string $transaction
     */
    public function __construct(Mailbox $mailbox, $transaction = '')
    {
        $this->mailbox = $mailbox;
        $this->transaction = $transaction;
    }

    /**
     * Execute the job.
     *
     * @return bool
     * @throws \Exception
     */
    public function handle()
    {
        $this->loadTheWorld();

        if ($this->mailboxIsBeingChecked()) {
            Log::debug('['.get_class($this).'] mailbox is being checked concurrently, bailing out', [
                'xMailbox' => $this->mailbox->xMailbox,
                'sUsername' => $this->mailbox->sUsername,
                'sHostname' => $this->mailbox->sHostname,
                'transaction' => $this->transaction,
            ]);
            return true;
        } else {
            $this->flagMailboxBeingChecked();
        }

        try {
            $box = new Imap();
            $box->connect($this->mailbox);
            Log::debug('['.get_class($this).'] connected to mailbox', [
                'xMailbox' => $this->mailbox->xMailbox,
                'sUsername' => $this->mailbox->sUsername,
                'sHostname' => $this->mailbox->sHostname,
                'transaction' => $this->transaction,
            ]);
        } catch(\Exception $e) {
            $this->removeMailboxBeingCheckedFlag();
            Log::error('Cant connect to #'. $this->mailbox->xMailbox .' - '. $e->getMessage());
        }


        // Bail out if we have no messages.
        if ($box->messageCount() == 0) {
            Log::debug('['.get_class($this).'] no messages found in mailbox', [
                'xMailbox' => $this->mailbox->xMailbox,
                'sUsername' => $this->mailbox->sUsername,
                'sHostname' => $this->mailbox->sHostname,
                'transaction' => $this->transaction,
            ]);
            $this->removeMailboxBeingCheckedFlag();
            return false;
        }

        // Save each to the disk and push onto parsing queue
        foreach ($box->getMessages() as $mid) {
            if($mid > cHD_EMAILS_MAX_TO_IMPORT){
                Log::info("[".get_class($this)."]  Stopped importing email, reached max retrieval limit (".cHD_EMAILS_MAX_TO_IMPORT.")", [
                    'xMailbox' => $this->mailbox->xMailbox,
                    'sUsername' => $this->mailbox->sUsername,
                    'sHostname' => $this->mailbox->sHostname,
                    'transaction' => $this->transaction,
                ]);
                break;
            }

            $fileName = date('Y-m-d-His').'-'.Str::uuid()->toString().'.eml.gz';
            $transaction = Str::uuid()->toString();

            try {
                $stored = $this->storeEml($fileName, $box->getMessage($mid));
            } catch (\Exception $e) {
                Log::error($e, [
                    'xMailbox' => $this->mailbox->xMailbox,
                    'filename' => $fileName,
                    'transaction' => $this->transaction,
                ]);
                continue;
            }

            if (! $stored) { // if it didn't store and no exception error out.
                Log::error('Failed to store eml', [
                    'xMailbox' => $this->mailbox->xMailbox,
                    'filename' => $fileName,
                    'transaction' => $this->transaction,
                ]);
                continue;
            }

            Log::debug('['.get_class($this).'] mail message eml stored, dispatching MailMessage', [
                'xMailbox' => $this->mailbox->xMailbox,
                'filename' => $fileName,
                'successful' => $stored,
                'mid' => $mid,
                'transaction' => $this->transaction,
                'child_transaction' => $transaction,
            ]);

            // Push onto queue to process
            MailMessage::dispatch($this->mailbox, $fileName, true, $transaction)
                ->onQueue(config('queue.high_priority_queue')); // mail.incoming

            // Mark as ready to delete from iMap server
            $expunged = $box->delete($mid);
            Log::debug('['.get_class($this).'] mail message marked for expunging', [
                'xMailbox' => $this->mailbox->xMailbox,
                'mid' => $mid,
                'successful' => $expunged,
                'transaction' => $this->transaction,
            ]);
        }

        $box->expunge();

        $this->removeMailboxBeingCheckedFlag();
        return true;
    }

    /**
     * @param $fileName
     * @param $contents
     * @return mixed
     */
    protected function storeEml($fileName, $contents)
    {
        return Storage::disk($this->storageDisk())
            ->put('mail/'.$fileName, gzencode($contents, 6));
    }

    public function visibleToAdministrators()
    {
        return true;
    }

    public function visibleMetaData()
    {
        return [
            'mailbox' => $this->mailbox->identify(),
            'transaction' => $this->transaction,
        ];
    }

    private function mailboxIsBeingChecked()
    {
        return Cache::has($this->mailboxCacheKey());
    }

    private function mailboxCacheKey()
    {
        return vsprintf('fetchmail:%s', [
            $this->mailbox->getKey()
        ]);
    }

    private function flagMailboxBeingChecked()
    {
        return Cache::put($this->mailboxCacheKey(), true, 60*10); // Set for 10 minutes
    }

    private function removeMailboxBeingCheckedFlag()
    {
        return Cache::forget($this->mailboxCacheKey());
    }
}
