<?php

use HS\Domain\Workspace\Document;
use HS\Domain\Workspace\History;
use Illuminate\Support\Facades\Schema;
use Illuminate\Database\Schema\Blueprint;
use Illuminate\Database\Migrations\Migration;

class AddHistoryIdToDocuments extends Migration
{
    /**
     * Run the migrations.
     *
     * @return void
     */
    public function up()
    {
        Schema::table('HS_Documents', function (Blueprint $table) {
            $table->integer('xRequestHistory')->nullable();
            $table->integer('xResponse')->nullable();
            $table->index('xRequestHistory');
        });

        $this->migrateHistoryDocuments();
        $this->migrateResponses();
        $this->dropDocumentIdColumnFromHistory();
    }

    /**
     * Reverse the migrations.
     *
     * @return void
     */
    public function down()
    {
        Schema::table('HS_Documents', function (Blueprint $table) {
            $table->dropColumn('xRequestHistory');
            $table->dropColumn('xResponse');
        });
    }

    /**
     * Loop through all the request history and
     * update the Document with the parent history id.
     */
    protected function migrateHistoryDocuments()
    {
        History::where('xDocumentId', '>', 0)
            ->orderBy('xRequestHistory', 'desc')
            ->chunk(1000, function($items){
                foreach ($items as $item) {
                    $parent = History::where('xDocumentId', '=', 0)
                        ->where('dtGMTChange', '=', $item->dtGMTChange)
                        ->first();

                    if ($parent) {
                        $doc = Document::find($item->xDocumentId);

                        if ($doc) {
                            $doc->xRequestHistory = $parent->xRequestHistory;
                            $doc->save();

                            // Delete the request history document artifact.
                            $item->delete();
                        }
                    }
                }
            });
    }

    /**
     * Loop through all the responses and
     * create the relationship with documents
     */
    protected function migrateResponses()
    {
        $responses = DB::table('HS_Responses')->get();
        foreach ($responses as $response) {
            $options = json_decode($response->tResponseOptions, true);
            if (isset($options['attachment']) && is_array($options['attachment'])) {
                foreach ($options['attachment'] as $attachId) {
                    $doc = Document::find($attachId);

                    if ($doc) {
                        $doc->xResponse = $response->xResponse;
                        $doc->save();
                    }
                }
            }
        }
    }

    /**
     * Drop the unused HS_Request_History.xDocumentId column
     */
    protected function dropDocumentIdColumnFromHistory()
    {
        if (config('database.default') == 'sqlsrv') {
            \HS\Database\SqlSrv::drop('HS_Request_History', 'xDocumentId');
        } else {
            Schema::table('HS_Request_History', function (Blueprint $table) {
                $table->dropColumn('xDocumentId');
            });
        }
    }
}
